from typing import Any, Callable, Optional, cast, get_origin

from pydantic import BaseModel, TypeAdapter
from pydantic.v1 import validate_arguments
from typing_extensions import Literal


def cast_callable_to_model(
    function: Callable[..., Any],
    name: Optional[str] = None,
    description: Optional[str] = None,
) -> type[BaseModel]:
    response = validate_arguments(function).model
    for field in ["args", "kwargs", "v__duplicate_kwargs"]:
        fields = cast(dict[str, Any], response.__fields__)
        fields.pop(field, None)
    response.__title__ = name or function.__name__
    response.__name__ = name or function.__name__
    response.__doc__ = description or function.__doc__
    return response


def parse_as(
    type_: Any,
    data: Any,
    mode: Literal["python", "json", "strings"] = "python",
) -> BaseModel:
    """Parse a json string to a Pydantic model."""
    adapter = TypeAdapter(type_)

    if get_origin(type_) is list and isinstance(data, dict):
        data = next(iter(data.values()))

    parser = getattr(adapter, f"validate_{mode}")

    return parser(data)



from typing import TYPE_CHECKING, Optional
from openai.types.beta.threads import ThreadMessage
from pydantic import BaseModel, Field

from cyberchipped.utilities.asyncio import (
    ExposeSyncMethodsMixin,
    expose_sync_method,
)
from cyberchipped.utilities.logging import get_logger
from cyberchipped.utilities.openai import get_client
from cyberchipped.utilities.pydantic import parse_as

logger = get_logger("Threads")

if TYPE_CHECKING:
    from .assistants import Assistant
    from .runs import Run


class Thread(BaseModel, ExposeSyncMethodsMixin):
    id: Optional[str] = None
    metadata: dict = {}
    messages: list[ThreadMessage] = Field([], repr=False)

    @expose_sync_method("get")
    async def get_async(self):
        """
        Gets a thread.
        """
        client = get_client()
        response = await client.beta.threads.retrieve(thread_id=self.id)
        self.id = response.id
        return self

    @expose_sync_method("create")
    async def create_async(self=None):
        """
        Creates a thread.
        """
        client = get_client()
        response = await client.beta.threads.create()
        self.id = response.id
        return self

    @expose_sync_method("add")
    async def add_async(
        self, message: str, file_paths: Optional[list[str]] = None
    ) -> ThreadMessage:
        """
        Add a user message to the thread.
        """
        client = get_client()

        # Create the message with the attached files
        response = await client.beta.threads.messages.create(
            thread_id=self.id,
            role="user",
            content=message,
        )
        return ThreadMessage.model_validate(response.model_dump())

    @expose_sync_method("get_messages")
    async def get_messages_async(
        self,
        limit: int = None,
        before_message: Optional[str] = None,
        after_message: Optional[str] = None,
    ):
        client = get_client()

        response = await client.beta.threads.messages.list(
            thread_id=self.id,
            # note that because messages are returned in descending order,
            # we reverse "before" and "after" to the API
            before=after_message,
            after=before_message,
            limit=limit,
            order="desc",
        )

        return parse_as(list[ThreadMessage], reversed(response.model_dump()["data"]))

    @expose_sync_method("delete")
    async def delete_async(self):
        client = get_client()
        await client.beta.threads.delete(thread_id=self.id)
        self.id = None

    @expose_sync_method("cancel_run")
    async def cancel_run_async(
        self,
        assistant: "Assistant",
        **run_kwargs,
    ) -> "Run":
        """
        Cancels the run of this thread with the provided assistant.
        """
        from cyberchipped.assistants.runs import Run

        run = Run(assistant=assistant, thread=self, **run_kwargs)
        return await run.cancel_async()

    @expose_sync_method("say")
    async def say_async(self, text: str, assistant: "Assistant") -> str:
        """
        Wraps the full process of adding a message to the thread and running it
        """
        from cyberchipped.assistants.runs import Run

        await self.add_async(text)

        try:
            await self.cancel_run_async(assistant=assistant)
        except Exception:
            pass

        run = Run(assistant=assistant, thread=self)
        await run.run_async()
        messages = await self.get_messages_async()
        last_message = messages[-1]
        ai_message = last_message.content[0].text.value
        return ai_message



import os
import pathlib
import shutil


def _safe_create_file(path: str) -> None:
    path = os.path.expanduser(path)
    file_path = pathlib.Path(path)
    file_path.parent.mkdir(parents=True, exist_ok=True)
    file_path.touch(exist_ok=True)


def getcwd() -> str:
    """Returns the current working directory"""
    return os.getcwd()


def write(path: str, contents: str) -> str:
    """Creates or overwrites a file with the given contents"""
    path = os.path.expanduser(path)
    _safe_create_file(path)
    with open(path, "w") as f:
        f.write(contents)
    return f'Successfully wrote "{path}"'


def write_lines(
    path: str, contents: str, insert_line: int = -1, mode: str = "insert"
) -> str:
    """Writes content to a specific line in the file.

    Args:
        path (str): The name of the file to write to.
        contents (str): The content to write to the file.
        insert_line (int, optional): The line number to insert the content at.
            Negative values count from the end of the file. Defaults to -1.
        mode (str, optional): The mode to use when writing the content. Can be
            "insert" or "overwrite". Defaults to "insert".

    Returns:
        str: A message indicating whether the write was successful.
    """
    path = os.path.expanduser(path)
    _safe_create_file(path)
    with open(path, "r") as f:
        lines = f.readlines()
        if insert_line < 0:
            insert_line = len(lines) + insert_line + 1
        if mode == "insert":
            lines[insert_line:insert_line] = contents.splitlines(True)
        elif mode == "overwrite":
            lines[insert_line : insert_line + len(contents.splitlines())] = (
                contents.splitlines(True)
            )
        else:
            raise ValueError(f"Invalid mode: {mode}")
    with open(path, "w") as f:
        f.writelines(lines)
    return f'Successfully wrote to "{path}"'


def read(path: str, include_line_numbers: bool = False) -> str:
    """Reads a file and returns the contents.

    Args:
        path (str): The path to the file.
        include_line_numbers (bool, optional): Whether to include line numbers
            in the returned contents. Defaults to False.

    Returns:
        str: The contents of the file.
    """
    path = os.path.expanduser(path)
    with open(path, "r") as f:
        if include_line_numbers:
            lines = f.readlines()
            lines_with_numbers = [f"{i+1}: {line}" for i, line in enumerate(lines)]
            return "".join(lines_with_numbers)
        else:
            return f.read()


def read_lines(
    path: str,
    start_line: int = 0,
    end_line: int = -1,
    include_line_numbers: bool = False,
) -> str:
    """Reads a partial file and returns the contents with optional line numbers.

    Args:
        path (str): The path to the file.
        start_line (int, optional): The starting line number to read. Defaults
            to 0.
        end_line (int, optional): The ending line number to read. Defaults to
            -1, which means read until the end of the file.
        include_line_numbers (bool, optional): Whether to include line numbers
            in the returned contents. Defaults to False.

    Returns:
        str: The contents of the file.
    """
    path = os.path.expanduser(path)
    with open(path, "r") as f:
        lines = f.readlines()
        if start_line < 0:
            start_line = len(lines) + start_line
        if end_line < 0:
            end_line = len(lines) + end_line
        if include_line_numbers:
            lines_with_numbers = [
                f"{i+1}: {line}" for i, line in enumerate(lines[start_line:end_line])
            ]
            return "".join(lines_with_numbers)
        else:
            return "".join(lines[start_line:end_line])


def mkdir(path: str) -> str:
    """Creates a directory (and any parent directories))"""
    path = os.path.expanduser(path)
    path = pathlib.Path(path)
    path.mkdir(parents=True, exist_ok=True)
    return f'Successfully created directory "{path}"'


def mv(src: str, dest: str) -> str:
    """Moves a file or directory"""
    src = os.path.expanduser(src)
    dest = os.path.expanduser(dest)
    src = pathlib.Path(src)
    dest = pathlib.Path(dest)
    src.rename(dest)
    return f'Successfully moved "{src}" to "{dest}"'


def cp(src: str, dest: str) -> str:
    """Copies a file or directory"""
    src = os.path.expanduser(src)
    dest = os.path.expanduser(dest)
    src = pathlib.Path(src)
    dest = pathlib.Path(dest)
    shutil.copytree(src, dest)
    return f'Successfully copied "{src}" to "{dest}"'


def ls(path: str) -> str:
    """Lists the contents of a directory"""
    path = os.path.expanduser(path)
    path = pathlib.Path(path)
    return "\n".join(str(p) for p in path.iterdir())



import inspect
from functools import partial
from typing import Any, Callable, Optional, TypeVar, Union, overload

from cyberchipped.components.ai_function import ai_fn
from cyberchipped.utilities.jinja import (
    BaseEnvironment,
)

T = TypeVar("T")

prompt = inspect.cleandoc(
    "The user will provide context as text that you need to parse into a structured"
    " form. To validate your response, you must call the"
    " `{{_response_model.function.name}}` function. Use the provided text to extract or"
    " infer any parameters needed by `{{_response_model.function.name}}`, including any"
    " missing data."
    " user: The text to parse: {{text}}"
)


@overload
def ai_model(
    *,
    environment: Optional[BaseEnvironment] = None,
    prompt: Optional[str] = prompt,
    model_name: str = "FormatResponse",
    model_description: str = "Formats the response.",
    field_name: str = "data",
    field_description: str = "The data to format.",
    **render_kwargs: Any,
) -> Callable[[T], Callable[[str], T]]:
    pass


@overload
def ai_model(
    _type: Optional[T],
    *,
    environment: Optional[BaseEnvironment] = None,
    prompt: Optional[str] = prompt,
    model_name: str = "FormatResponse",
    model_description: str = "Formats the response.",
    field_name: str = "data",
    field_description: str = "The data to format.",
    **render_kwargs: Any,
) -> Callable[[str], T]:
    pass


def ai_model(
    _type: Optional[T] = None,
    *,
    environment: Optional[BaseEnvironment] = None,
    prompt: Optional[str] = prompt,
    model_name: str = "FormatResponse",
    model_description: str = "Formats the response.",
    field_name: str = "data",
    field_description: str = "The data to format.",
    **render_kwargs: Any,
) -> Union[Callable[[T], Callable[[str], T]], Callable[[str], T],]:
    def wrapper(_type_: T, text: str) -> T:
        def extract(text: str) -> T:  # type: ignore
            pass

        extract.__annotations__["return"] = _type_

        return ai_fn(
            extract,
            environment=environment,
            prompt=prompt,
            model_name=model_name,
            model_description=model_description,
            field_name=field_name,
            field_description=field_description,
            **render_kwargs,
        )(text)

    if _type is not None:
        return partial(wrapper, _type)

    def decorator(_type_: T) -> Callable[[str], T]:
        return partial(wrapper, _type)

    return decorator



from .ai_function import ai_fn, AIFunction
from .ai_model import ai_model
from .prompt import prompt_fn, PromptFunction

__all__ = [
    "ai_fn",
    "ai_model",
    "prompt_fn",
    "AIFunction",
    "PromptFunction",
]



import asyncio
import os
from functools import lru_cache
from typing import Optional

from openai import AsyncClient


def get_client() -> AsyncClient:
    from cyberchipped import settings

    api_key: Optional[str] = os.getenv("OPENAI_API_KEY")
    if not api_key:
        api_key = (
            settings.openai.api_key.get_secret_value()
            if settings.openai.api_key
            else None
        )
    organization: Optional[str] = settings.openai.organization
    return _get_client_memoized(
        api_key=api_key, organization=organization, loop=asyncio.get_event_loop()
    )


@lru_cache
def _get_client_memoized(
    api_key: Optional[str],
    organization: Optional[str],
    loop: Optional[asyncio.AbstractEventLoop] = None,
) -> AsyncClient:
    """
    This function is memoized to ensure that only one instance of the client is
    created for a given api key / organization / loop tuple.

    The `loop` is an important key to ensure that the client is not re-used
    across multiple event loops (which can happen when using the `run_sync`
    function). Attempting to re-use the client across multiple event loops
    can result in a `RuntimeError: Event loop is closed` error or infinite hangs.
    """
    return AsyncClient(
        api_key=api_key,
        organization=organization,
    )



from typing import TYPE_CHECKING, Callable, Optional, Union

from pydantic import BaseModel, Field, field_validator

import cyberchipped.utilities.tools
from cyberchipped.requests import Tool
from cyberchipped.tools.assistants import AssistantTools
from cyberchipped.utilities.asyncio import (
    ExposeSyncMethodsMixin,
    expose_sync_method,
    run_sync,
)
from cyberchipped.utilities.logging import get_logger
from cyberchipped.utilities.openai import get_client

from .threads import Thread

from dotenv import load_dotenv
import os

load_dotenv(dotenv_path=os.path.join(os.getcwd(), ".env"))

if TYPE_CHECKING:
    from .runs import Run

logger = get_logger("Assistants")


class Assistant(BaseModel, ExposeSyncMethodsMixin):
    id: Optional[str] = None
    name: str = "Assistant"
    model: str = "gpt-4-1106-preview"
    instructions: Optional[str] = Field(None, repr=False)
    tools: list[AssistantTools] = []
    metadata: dict[str, str] = {}

    def get_tools(self) -> list[AssistantTools]:
        return self.tools

    def get_instructions(self) -> str:
        return self.instructions or ""

    @field_validator("tools", mode="before")
    def format_tools(cls, tools: list[Union[Tool, Callable]]):
        return [
            (
                tool
                if isinstance(tool, Tool)
                else cyberchipped.utilities.tools.tool_from_function(tool)
            )
            for tool in tools
        ]

    def __enter__(self):
        self.create()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.delete()
        # If an exception has occurred, you might want to handle it or pass it through
        # Returning False here will re-raise any exception that occurred in the context
        return False

    @expose_sync_method("create")
    async def create_async(self):
        client = get_client()
        response = await client.beta.assistants.create(
            **self.model_dump(include={"name", "model", "metadata", "metadata"}),
            tools=[tool.model_dump() for tool in self.get_tools()],
            instructions=self.get_instructions(),
        )
        self.id = response.id

    @expose_sync_method("delete")
    async def delete_async(self):
        if not self.id:
            raise ValueError("Assistant has not been created.")
        client = get_client()
        await client.beta.assistants.delete(assistant_id=self.id)
        self.id = None

    @staticmethod
    def get_db_connection():
        import sqlite3

        conn = sqlite3.connect("cyberchipped.db")
        c = conn.cursor()
        c.execute(
            """
            CREATE TABLE IF NOT EXISTS threads
            (id TEXT PRIMARY KEY, assistant_name TEXT, user_id TEXT)
        """
        )
        return conn, c

    @staticmethod
    def get_mongo_connection():
        from pymongo import MongoClient

        client = MongoClient(os.getenv("MONGO_URL"))
        db = client[os.getenv("MONGO_DB")][os.getenv("MONGO_COLLECTION")]
        return db

    @staticmethod
    def should_use_mongo() -> bool:
        return os.getenv("MONGO_URL") is not None

    @expose_sync_method("say")
    async def say_async(self, text: str, user_id=None) -> "Run":
        if self.should_use_mongo():
            db = self.get_mongo_connection()
            if user_id is None:
                thread = db.find_one({"assistant_name": self.name})
            else:
                thread = db.find_one({"user_id": user_id, "assistant_name": self.name})

            if thread is None:
                thread = Thread()
                await thread.create_async()
                if not user_id:
                    db.insert_one(
                        {"id": thread.id, "assistant_name": self.name, "user_id": ""}
                    )
                else:
                    db.insert_one(
                        {
                            "id": thread.id,
                            "assistant_name": self.name,
                            "user_id": user_id,
                        }
                    )
            else:
                thread_id = thread["id"]
                thread = Thread(id=thread_id)
                await thread.get_async()

        else:
            conn, c = self.get_db_connection()
            if user_id is None:
                c.execute(
                    "SELECT id FROM threads WHERE assistant_name = ? LIMIT 1",
                    (self.name,),
                )
            else:
                c.execute(
                    "SELECT id FROM threads WHERE user_id = ? AND assistant_name = ? LIMIT 1",
                    (user_id, self.name),
                )

            result = c.fetchone()

            if result is None:
                thread = Thread()
                await thread.create_async()
                if not user_id:
                    c.execute(
                        "INSERT INTO threads VALUES (?, ?, ?)",
                        (thread.id, self.name, ""),
                    )
                else:
                    c.execute(
                        "INSERT INTO threads VALUES (?, ?, ?)",
                        (thread.id, self.name, user_id),
                    )
                conn.commit()
            else:
                thread_id = result[0]
                thread = Thread(id=thread_id)
                await thread.get_async()

            conn.close()

        return await thread.say_async(text, assistant=self)

    @expose_sync_method("get_default_thread")
    async def get_default_thread_async(self) -> "Thread":
        if self.should_use_mongo():
            db = self.get_mongo_connection()
            thread = db.find_one({"assistant_name": self.name})
            thread_id = thread["id"]
        else:
            conn, c = Assistant.get_db_connection()
            c.execute(
                "SELECT id FROM threads WHERE assistant_name = ? LIMIT 1", (self.name,)
            )
            result = c.fetchone()
            thread_id = result[0]
            conn.close()

        thread = Thread(id=thread_id)
        await thread.get_async()

        return thread

    @classmethod
    def load(cls, assistant_id: str):
        return run_sync(cls.load_async(assistant_id))

    @classmethod
    async def load_async(cls, assistant_id: str):
        client = get_client()
        response = await client.beta.assistants.retrieve(assistant_id=assistant_id)
        return cls.model_validate(response)



from datetime import datetime
from openai.types.beta.threads import ThreadMessage
from rich import box
from rich.console import Console
from rich.panel import Panel


def pprint_message(message: ThreadMessage):
    """
    Pretty-prints a single message using the rich library, highlighting the
    speaker's role, the message text, any available images, and the message
    timestamp in a panel format.

    Args:
    message (dict): A message object as described in the API documentation.
    """
    console = Console()
    role_colors = {
        "user": "green",
        "assistant": "blue",
    }

    color = role_colors.get(message.role, "red")
    timestamp = datetime.fromtimestamp(message.created_at).strftime("%l:%M:%S %p")

    content = ""
    for item in message.content:
        if item.type == "text":
            content += item.text.value + "\n\n"

    # Create the panel for the message
    panel = Panel(
        content.strip(),
        title=f"[bold]{message.role.capitalize()}[/]",
        subtitle=f"[italic]{timestamp}[/]",
        title_align="left",
        subtitle_align="right",
        border_style=color,
        box=box.ROUNDED,
        # highlight=True,
        width=100,  # Fixed width for all panels
        expand=True,  # Panels always expand to the width of the console
        padding=(1, 2),
    )

    # Printing the panel
    console.print(panel)


def pprint_messages(messages: list[ThreadMessage]):
    for message in messages:
        pprint_message(message)



# ruff: noqa
from .runs import Run
from .threads import Thread
from .assistants import Assistant
from .formatting import pprint_message, pprint_messages
from cyberchipped.tools.assistants import Retrieval, CodeInterpreter



import asyncio
from typing import Any, Callable, Optional, Union

from openai.types.beta.threads.run import Run as OpenAIRun
from pydantic import BaseModel, Field, field_validator

import cyberchipped.utilities.tools
from cyberchipped.requests import Tool
from cyberchipped.tools.assistants import AssistantTools, CancelRun
from cyberchipped.utilities.asyncio import ExposeSyncMethodsMixin, expose_sync_method
from cyberchipped.utilities.logging import get_logger
from cyberchipped.utilities.openai import get_client

from .assistants import Assistant
from .threads import Thread

logger = get_logger("Runs")


class Run(BaseModel, ExposeSyncMethodsMixin):
    thread: Thread
    assistant: Assistant
    instructions: Optional[str] = Field(
        None, description="Replacement instructions to use for the run."
    )
    additional_instructions: Optional[str] = Field(
        None,
        description=(
            "Additional instructions to append to the assistant's instructions."
        ),
    )
    tools: Optional[list[Union[AssistantTools, Callable]]] = Field(
        None, description="Replacement tools to use for the run."
    )
    additional_tools: Optional[list[AssistantTools]] = Field(
        None,
        description="Additional tools to append to the assistant's tools. ",
    )
    run: OpenAIRun = None
    data: Any = None

    @field_validator("tools", "additional_tools", mode="before")
    def format_tools(cls, tools: Union[None, list[Union[Tool, Callable]]]):
        if tools is not None:
            return [
                (
                    tool
                    if isinstance(tool, Tool)
                    else cyberchipped.utilities.tools.tool_from_function(tool)
                )
                for tool in tools
            ]

    @expose_sync_method("refresh")
    async def refresh_async(self):
        client = get_client()
        self.run = await client.beta.threads.runs.retrieve(
            run_id=self.run.id, thread_id=self.thread.id
        )

    @expose_sync_method("cancel")
    async def cancel_async(self):
        client = get_client()
        await client.beta.threads.runs.cancel(
            run_id=self.run.id, thread_id=self.thread.id
        )

    async def _handle_step_requires_action(self):
        client = get_client()
        if self.run.status != "requires_action":
            return
        if self.run.required_action.type == "submit_tool_outputs":
            tool_outputs = []
            tools = self.get_tools()

            for tool_call in self.run.required_action.submit_tool_outputs.tool_calls:
                try:
                    output = cyberchipped.utilities.tools.call_function_tool(
                        tools=tools,
                        function_name=tool_call.function.name,
                        function_arguments_json=tool_call.function.arguments,
                    )
                except CancelRun as exc:
                    logger.debug(f"Ending run with data: {exc.data}")
                    raise
                except Exception as exc:
                    output = f"Error calling function {tool_call.function.name}: {exc}"
                    logger.error(output)
                tool_outputs.append(
                    dict(tool_call_id=tool_call.id, output=output or "")
                )

            await client.beta.threads.runs.submit_tool_outputs(
                thread_id=self.thread.id, run_id=self.run.id, tool_outputs=tool_outputs
            )

    def get_instructions(self) -> str:
        if self.instructions is None:
            instructions = self.assistant.get_instructions() or ""
        else:
            instructions = self.instructions

        if self.additional_instructions is not None:
            instructions = "\n\n".join([instructions, self.additional_instructions])

        return instructions

    def get_tools(self) -> list[AssistantTools]:
        tools = []
        if self.tools is None:
            tools.extend(self.assistant.get_tools())
        else:
            tools.extend(self.tools)
        if self.additional_tools is not None:
            tools.extend(self.additional_tools)
        return tools

    async def run_async(self) -> "Run":
        client = get_client()

        create_kwargs = {}

        if self.instructions is not None or self.additional_instructions is not None:
            create_kwargs["instructions"] = self.get_instructions()

        if self.tools is not None or self.additional_tools is not None:
            create_kwargs["tools"] = self.get_tools()

        self.run = await client.beta.threads.runs.create(
            thread_id=self.thread.id, assistant_id=self.assistant.id, **create_kwargs
        )

        try:
            await asyncio.wait_for(self._run_loop(), timeout=60)
        except asyncio.TimeoutError:
            if self.run.status != "completed":
                # Cancel the run if it's not completed
                await client.beta.threads.runs.cancel(
                    run_id=self.run.id, thread_id=self.thread.id
                )
                self.data = "Run cancelled due to timeout."
            else:
                self.data = "Run already completed; no need to cancel."
            await self.refresh_async()

    async def _run_loop(self):
        while self.run.status in ("queued", "in_progress", "requires_action"):
            if self.run.status == "requires_action":
                await self._handle_step_requires_action()
            await asyncio.sleep(0.1)
            await self.refresh_async()



import inspect
import re
from functools import partial, wraps
from typing import (
    Any,
    Callable,
    Optional,
    TypeVar,
    Union,
    overload,
)

import pydantic
from pydantic import BaseModel
from typing_extensions import ParamSpec, Self

from cyberchipped.requests import BaseMessage as Message
from cyberchipped.requests import Prompt, Tool
from cyberchipped.serializers import (
    create_grammar_from_vocabulary,
    create_tool_from_type,
    create_vocabulary_from_type,
)
from cyberchipped.settings import settings
from cyberchipped.utilities.jinja import (
    BaseEnvironment,
    Transcript,
)

P = ParamSpec("P")
T = TypeVar("T")
U = TypeVar("U", bound=BaseModel)


class PromptFunction(Prompt[U]):
    messages: list[Message] = pydantic.Field(default_factory=list)

    def serialize(self) -> dict[str, Any]:
        return self.model_dump(exclude_unset=True, exclude_none=True)

    @overload
    @classmethod
    def as_grammar(
        cls: type[Self],
        *,
        environment: Optional[BaseEnvironment] = None,
        prompt: Optional[str] = None,
        enumerate: bool = True,
        encoder: Callable[[str], list[int]] = settings.openai.chat.completions.encoder,
        max_tokens: Optional[int] = 1,
    ) -> Callable[[Callable[P, Any]], Callable[P, Self]]:
        pass

    @overload
    @classmethod
    def as_grammar(
        cls: type[Self],
        fn: Optional[Callable[P, Any]] = None,
        *,
        environment: Optional[BaseEnvironment] = None,
        prompt: Optional[str] = None,
        enumerate: bool = True,
        encoder: Callable[[str], list[int]] = settings.openai.chat.completions.encoder,
        max_tokens: Optional[int] = 1,
    ) -> Callable[P, Self]:
        pass

    @classmethod
    def as_grammar(
        cls: type[Self],
        fn: Optional[Callable[P, Any]] = None,
        *,
        environment: Optional[BaseEnvironment] = None,
        prompt: Optional[str] = None,
        enumerate: bool = True,
        encoder: Callable[[str], list[int]] = settings.openai.chat.completions.encoder,
        max_tokens: Optional[int] = 1,
        **kwargs: Any,
    ) -> Union[Callable[[Callable[P, Any]], Callable[P, Self]], Callable[P, Self],]:
        def wrapper(func: Callable[P, Any], *args: P.args, **kwargs: P.kwargs) -> Self:
            # Get the signature of the function
            signature = inspect.signature(func)
            params = signature.bind(*args, **kwargs)
            params.apply_defaults()

            vocabulary = create_vocabulary_from_type(
                inspect.signature(func).return_annotation
            )

            grammar = create_grammar_from_vocabulary(
                vocabulary=vocabulary,
                encoder=encoder,
                _enumerate=enumerate,
                max_tokens=max_tokens,
            )

            messages = Transcript(
                content=prompt or func.__doc__ or ""
            ).render_to_messages(
                **kwargs | params.arguments,
                _arguments=params.arguments,
                _options=vocabulary,
                _doc=func.__doc__,
                _source_code=(
                    "\ndef" + "def".join(re.split("def", inspect.getsource(func))[1:])
                ),
            )

            return cls(
                messages=messages,
                **grammar.model_dump(exclude_unset=True, exclude_none=True),
            )

        if fn is not None:
            return wraps(fn)(partial(wrapper, fn))

        def decorator(fn: Callable[P, Any]) -> Callable[P, Self]:
            return wraps(fn)(partial(wrapper, fn))

        return decorator

    @overload
    @classmethod
    def as_function_call(
        cls: type[Self],
        *,
        environment: Optional[BaseEnvironment] = None,
        prompt: Optional[str] = None,
        model_name: str = "FormatResponse",
        model_description: str = "Formats the response.",
        field_name: str = "data",
        field_description: str = "The data to format.",
    ) -> Callable[[Callable[P, Any]], Callable[P, Self]]:
        pass

    @overload
    @classmethod
    def as_function_call(
        cls: type[Self],
        fn: Optional[Callable[P, Any]] = None,
        *,
        environment: Optional[BaseEnvironment] = None,
        prompt: Optional[str] = None,
        model_name: str = "FormatResponse",
        model_description: str = "Formats the response.",
        field_name: str = "data",
        field_description: str = "The data to format.",
    ) -> Callable[P, Self]:
        pass

    @classmethod
    def as_function_call(
        cls: type[Self],
        fn: Optional[Callable[P, Any]] = None,
        *,
        environment: Optional[BaseEnvironment] = None,
        prompt: Optional[str] = None,
        model_name: str = "FormatResponse",
        model_description: str = "Formats the response.",
        field_name: str = "data",
        field_description: str = "The data to format.",
        **kwargs: Any,
    ) -> Union[Callable[[Callable[P, Any]], Callable[P, Self]], Callable[P, Self],]:
        def wrapper(func: Callable[P, Any], *args: P.args, **kwargs: P.kwargs) -> Self:
            # Get the signature of the function
            signature = inspect.signature(func)
            params = signature.bind(*args, **kwargs)
            params.apply_defaults()

            tool = create_tool_from_type(
                _type=inspect.signature(func).return_annotation,
                model_name=model_name,
                model_description=model_description,
                field_name=field_name,
                field_description=field_description,
            )

            messages = Transcript(
                content=prompt or func.__doc__ or ""
            ).render_to_messages(
                **kwargs | params.arguments,
                _doc=func.__doc__,
                _arguments=params.arguments,
                _response_model=tool,
                _source_code=(
                    "\ndef" + "def".join(re.split("def", inspect.getsource(func))[1:])
                ),
            )

            return cls(
                messages=messages,
                tool_choice={
                    "type": "function",
                    "function": {"name": getattr(tool.function, "name", model_name)},
                },
                tools=[Tool[BaseModel](**tool.model_dump())],
            )

        if fn is not None:
            return wraps(fn)(partial(wrapper, fn))

        def decorator(fn: Callable[P, Any]) -> Callable[P, Self]:
            return wraps(fn)(partial(wrapper, fn))

        return decorator


def prompt_fn(
    fn: Optional[Callable[P, T]] = None,
    *,
    environment: Optional[BaseEnvironment] = None,
    prompt: Optional[str] = None,
    model_name: str = "FormatResponse",
    model_description: str = "Formats the response.",
    field_name: str = "data",
    field_description: str = "The data to format.",
    **kwargs: Any,
) -> Union[
    Callable[[Callable[P, T]], Callable[P, dict[str, Any]]],
    Callable[P, dict[str, Any]],
]:
    def wrapper(
        func: Callable[P, Any], *args: P.args, **kwargs: P.kwargs
    ) -> dict[str, Any]:
        return PromptFunction.as_function_call(
            fn=func,
            environment=environment,
            prompt=prompt,
            model_name=model_name,
            model_description=model_description,
            field_name=field_name,
            field_description=field_description,
            **kwargs,
        )(*args, **kwargs).serialize()

    if fn is not None:
        return wraps(fn)(partial(wrapper, fn))

    def decorator(fn: Callable[P, Any]) -> Callable[P, dict[str, Any]]:
        return wraps(fn)(partial(wrapper, fn))

    return decorator



import contextvars
from contextlib import contextmanager


class ScopedContext:
    def __init__(self):
        self._context_storage = contextvars.ContextVar(
            "scoped_context_storage", default={}
        )

    def get(self, key, default=None):
        return self._context_storage.get().get(key, default)

    def set(self, **kwargs):
        ctx = self._context_storage.get()
        updated_ctx = {**ctx, **kwargs}
        self._context_storage.set(updated_ctx)

    @contextmanager
    def __call__(self, **kwargs):
        current_context = self._context_storage.get().copy()
        self.set(**kwargs)
        try:
            yield
        finally:
            self._context_storage.set(current_context)
